@@ -16,19 +16,20 @@
 # You should have received a copy of the GNU General Public License
 # along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
 
-import os, sys, shutil, urllib
+from __future__ import print_function
+import os, sys, shutil
 import json
-from ligo.gracedb.rest import GraceDb, load_json_or_die, GraceDbBasic
+from ligo.gracedb.rest import GraceDb, GraceDbBasic
+import six
 
-
-DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/gracedb/api"
+DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api"
 
 DEFAULT_BASIC_URL = "https://gracedb.ligo.org/apibasic/"
 
-GIT_TAG = 'gracedb-1.20-1'
+GIT_TAG = 'gracedb-1.27-1'
 
 DEFAULT_COLUMNS = "graceid,labels,group,pipeline,search,far,gpstime,created,dataurl"
- 
+
 #-----------------------------------------------------------------
 # Util routines
 
@@ -74,11 +75,11 @@ typeCodeMap = {
         "MBTAOnline": "MBTA",
         "Injection": "HWINJ",
 }
-validTypes = typeCodeMap.keys()
+validTypes = list(typeCodeMap.keys())
 
 #-----------------------------------------------------------------
-# This is a factory for client classes. 
-# Given a base client class with the correct properties, derive 
+# This is a factory for client classes.
+# Given a base client class with the correct properties, derive
 # one suitable for use with this command-line tool. In practice,
 # the base class here will either be the X509 auth GraceDb class
 # or the basic auth GraceDbBasic class.
@@ -95,16 +96,16 @@ def derive_client(ClientBase=GraceDb):
             # Check that we *could* write the file before we
             # go to the trouble of getting it.  Also, try not
             # to open a file until we know we have data.
-            if not isinstance(destfile, file) and destfile != "-":
+            if not hasattr(destfile, 'read') and destfile != "-":
                 if not os.access(os.path.dirname(os.path.abspath(destfile)), os.W_OK):
                     raise IOError("%s: Permission denied" % destfile)
             response = self.files(graceid, filename)
             if response.status == 200:
-                if not isinstance(destfile, file):
+                if not hasattr(destfile, 'read'):
                     if destfile == '-':
                         destfile = sys.stdout
                     else:
-                        destfile = open(destfile, "w")
+                        destfile = open(destfile, "wb")
                 shutil.copyfileobj(response, destfile)
                 return 0
             else:
@@ -113,8 +114,14 @@ def derive_client(ClientBase=GraceDb):
         # Hamstring 'adjustResponse' from the example REST client.
         # We don't want it messing with the response from the server.
         def adjustResponse(self, response):
-            response.json = lambda: load_json_or_die(response)
+            response.json = lambda: self.load_json_or_die(response)
             return response
+
+        @classmethod
+        def output_and_die(cls, msg):
+            sys.stderr.write(msg)
+            sys.exit(1)
+
     return client
 
 # Get X509 client in global scope
@@ -122,12 +129,12 @@ def derive_client(ClientBase=GraceDb):
 Client = derive_client()
 
 #-----------------------------------------------------------------
-# Main 
+# Main
 
 
 def main():
     usage ="""%%prog [options] GROUP PIPELINE SEARCH EVENTFILE
-   where GROUP is one of %(groups)s
+    where GROUP is one of %(groups)s
          PIPELINE is one of %(pipelines)s
          SEARCH (optional) is one of %(searches)s
          EVENTFILE is file containing event data. '-' indicates stdin.
@@ -145,11 +152,11 @@ def main():
 %%prog [options] ping
    Test server connection
 
-%%prog [options] upload GRACEID FILE [COMMENT] 
+%%prog [options] upload GRACEID FILE [COMMENT]
    where GRACEID is the id of an existing candidate event in GraCEDb
          FILE      is the name of the file to upload. '-' indicates stdin.
          COMMENT   is an optional annotation to enter into the log
-   Upload FILE to the private data area for a candidate event. To apply 
+   Upload FILE to the private data area for a candidate event. To apply
    a tag, use the --tag-name option (and --tag-display-name if desired.)
 
 %%prog [options] download GRACEID FILE [DESTINATION]
@@ -173,12 +180,12 @@ def main():
          LOG_N     is the number of the log message.
          TAG_NAME  is the name of the tag
          DISP_NAME is the tag display name (ignored for existing tags)
-    Tag an existing log message. Alternatively, the tag name and 
-    display name can be passed in with the --tag-name and 
+    Tag an existing log message. Alternatively, the tag name and
+    display name can be passed in with the --tag-name and
     --tag-display-name options.
 
 %%prog [options] delete_tag GRACEID LOG_N TAG_NAME
-    Remove a tag from a log message. Alternatively, the tag name 
+    Remove a tag from a log message. Alternatively, the tag name
     can be passed in with the --tag-name option.
 
 %%prog [options] search SEARCH PARAMS
@@ -245,8 +252,18 @@ Longer strings will be truncated.""" % {
                   help="Use basic auth with a .netrc file. Available to non-LVC members only.",
                   action="store_true", default=False
                  )
+    op.add_option("-o", "--offline", dest="offline", action="store_true",
+                  help=("Signifies that an event was found by an offline "
+                  "pipeline. Used when creating a new event."), default=False
+                 )
+    op.add_option("--labels", dest="labels",help=("Defines labels for creating"
+                  " a new event. Should be a comma-separated list of labels."),
+                  default=[], type=str
+                 )
 
     options, args = op.parse_args()
+    if isinstance(options.labels, str):
+        options.labels = options.labels.split(',')
 
     try:
         from glue.ligolw import ligolw
@@ -269,7 +286,7 @@ Longer strings will be truncated.""" % {
 
     # If the user requested a specific service, but also wants basic auth,
     # then the service had better be a basic auth endpoint. Otherwise die.
-    # On the other hand, if the user did not specify a service url, then we 
+    # On the other hand, if the user did not specify a service url, then we
     # will use the default basic URL if basic auth was requested.
     if options.use_basic_auth:
         if options.service or os.environ.get('GRACEDB_SERVICE_URL', None):
@@ -279,7 +296,7 @@ Longer strings will be truncated.""" % {
         else:
             service = DEFAULT_BASIC_URL
 
-    # Client subclass according to preferred auth method. 
+    # Client subclass according to preferred auth method.
     global Client
     if options.use_basic_auth:
         Client = derive_client(GraceDbBasic)
@@ -306,7 +323,7 @@ Longer strings will be truncated.""" % {
     elif args[0] == 'version':
         import pkg_resources
         version = pkg_resources.require("ligo-gracedb")[0].version
-        print "GraceDB Client v. %s" % version
+        print("GraceDB Client v. %s" % version)
         exit(0)
     elif args[0] == 'ping':
         response = client.ping()
@@ -360,7 +377,7 @@ Longer strings will be truncated.""" % {
         response = client.download(graceid, filename, outfile)
         if response:
             # no response means file saved.  any other response is an error message.
-            print response
+            print(response)
             exit(1)
         exit(0)
     elif args[0] == 'log':
@@ -407,7 +424,7 @@ Longer strings will be truncated.""" % {
         response = client.writeLabel(graceid, label)
     elif args[0] == 'search':
         query = " ".join(args[1:])
-        
+
         columns = options.columns
         columns = columns.replace('DEFAULTS',DEFAULT_COLUMNS)
         columns = columns.split(',')
@@ -434,7 +451,7 @@ Longer strings will be truncated.""" % {
         else:
             accessFun = {
                 "labels" : lambda e: \
-                    ",".join(e['labels'].keys()),
+                    ",".join(list(e['labels'].keys())),
                 "dataurl" : lambda e: e['links']['files'],
             }
 
@@ -444,7 +461,7 @@ Longer strings will be truncated.""" % {
                 row = [ accessFun.get(column, lambda e: defaultAccess(e,column))(e) for column in columns ]
                 row = "\t".join(row)
                 output(row)
-        
+
         return 0
     elif args[0] == 'replace':
         if len(args) != 3:
@@ -464,23 +481,25 @@ Longer strings will be truncated.""" % {
             filename = args[3]
 
         # Check that the group, search, and pipeline are known to the API.
-        foundGroup = True if (unicode(group) in client.groups) else False
+        foundGroup = True if (six.text_type(group) in client.groups) else False
         if not foundGroup:
             error("Group must be one of: ", ", ".join(client.groups))
             sys.exit(1)
 
-        foundPipeline = True if (unicode(pipeline) in client.pipelines) else False
+        foundPipeline = True if (six.text_type(pipeline) in client.pipelines) else False
         if not foundPipeline:
             error("Pipeline must be one of: ", ", ".join(client.pipelines))
             sys.exit(1)
 
         if search:
-            foundSearch = True if (unicode(search) in client.searches) else False
+            foundSearch = True if (six.text_type(search) in client.searches) else False
             if not foundSearch:
                 error("Search must be one of: ", ", ".join(client.searches))
                 sys.exit(1)
 
-        response = client.createEvent(group, pipeline, filename, search)
+        response = client.createEvent(group, pipeline, filename, search,
+            offline=options.offline, labels=options.labels)
+
         if not response:
             error("There was a problem.  Did you do ligo-proxy-init?")
             sys.exit(1)
@@ -500,9 +519,9 @@ Longer strings will be truncated.""" % {
         except:
             pass
 
-        if 'graceid' in rv.keys():
+        if 'graceid' in list(rv.keys()):
             output(rv['graceid'])
-        elif 'error' in rv.keys():
+        elif 'error' in list(rv.keys()):
             exitCode=1
             error(rv['error'])
 
