@@ -16,17 +16,23 @@
 # You should have received a copy of the GNU General Public License
 # along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
 
-import httplib, socket, ssl
+import six.moves.http_client, socket, ssl
 import mimetypes
 import urllib
-import os, sys
+import os
+import sys
+if os.name == 'posix':
+    import pwd
 import json
-from urlparse import urlparse
+from six.moves.urllib.parse import urlparse
 from base64 import b64encode
-import netrc
+from netrc import netrc, NetrcParseError
 from subprocess import Popen, PIPE
 import shlex
+import stat
 from datetime import datetime
+import six
+from six.moves import map
 
 DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api/"
 DEFAULT_BASIC_SERVICE_URL = "https://gracedb.ligo.org/apibasic/"
@@ -52,19 +58,50 @@ if sys.hexversion < 0x20709f0:
     ssl.wrap_socket = wrap_socket_patched
 
 #-----------------------------------------------------------------
-# Utilities 
+# Utilities
+
+class safe_netrc(netrc):
+    """The netrc.netrc class from the Python standard library applies access
+    safety checks (requiring that the netrc file is readable only by the
+    current user, and not by group members or other users) only if using the
+    netrc file in the default location (~/.netrc). This subclass applies the
+    same access safety checks regardless of the path to the netrc file."""
+
+    def _parse(self, file, fp, *args, **kwargs):
+        # Copied and adapted from netrc.py from Python 2.7
+        if os.name == 'posix':
+            prop = os.fstat(fp.fileno())
+            if prop.st_uid != os.getuid():
+                try:
+                    fowner = pwd.getpwuid(prop.st_uid)[0]
+                except KeyError:
+                    fowner = 'uid %s' % prop.st_uid
+                try:
+                    user = pwd.getpwuid(os.getuid())[0]
+                except KeyError:
+                    user = 'uid %s' % os.getuid()
+                raise NetrcParseError(
+                    ("~/.netrc file owner (%s) does not match"
+                     " current user (%s)") % (fowner, user),
+                    file)
+            if (prop.st_mode & (stat.S_IRWXG | stat.S_IRWXO)):
+                raise NetrcParseError(
+                   "~/.netrc access too permissive: access"
+                   " permissions must restrict access to only"
+                   " the owner", file)
+        return netrc._parse(self, file, fp, *args, **kwargs)
 
 # XXX It would be nice to get rid of this if we can.
 # It seems that you can pass python lists via the requests package.
-# That would make putting our lists into comma-separated strings 
+# That would make putting our lists into comma-separated strings
 # unnecessary.
 def cleanListInput(list_arg):
     stringified_list = list_arg
     if isinstance(list_arg, float) or isinstance(list_arg, int):
         stringified_value = str(list_arg)
         return stringified_value
-    if not isinstance(list_arg, basestring):
-        stringified_list = ','.join(map(str,list_arg))            
+    if not isinstance(list_arg, six.string_types):
+        stringified_list = ','.join(map(str,list_arg))
     return stringified_list
 
 # The following are used to check whether a user has tried to use
@@ -109,40 +146,6 @@ def is_expired(cert_file):
 
     return expired, err
 
-# A utility for writing out an error message to the user and then stopping
-# execution. This seems to behave sensibly in both the interpreter and in
-# a script.
-def output_and_die(msg):
-    sys.stderr.write(msg)
-    sys.exit(1)
-
-# Given an HTTPResponse object, try to read it's content and interpret as
-# JSON--or die trying.
-def load_json_or_die(response):
-
-    # First check that the response object actually exists.
-    if not response:
-        msg = "ERROR: no response object. \n\n"
-        output_and_die(msg)
-
-    # Next, try to read the content of the response.
-    try:
-        response_content = response.read()
-    except Exception, e:
-        msg = "ERROR: problem reading response. \n\n"
-        output_and_die(msg)
-
-    # Finally, try to create a dict by decoding the response as JSON.
-    rdict = None
-    try:
-        rdict = json.loads(response_content)
-    except Exception, e:
-        msg = "ERROR: got unexpected content from the server:\n\n"
-        msg += response_content + "\n\n"
-        output_and_die(msg)
-        
-    return rdict          
-
 #-----------------------------------------------------------------
 # Exception(s)
 
@@ -157,7 +160,7 @@ class HTTPError(Exception):
 # HTTP/S Proxy classes
 # Taken from: http://code.activestate.com/recipes/456195/
 
-class ProxyHTTPConnection(httplib.HTTPConnection):
+class ProxyHTTPConnection(six.moves.http_client.HTTPConnection):
 
     _ports = {'http' : 80, 'https' : 443}
 
@@ -169,19 +172,19 @@ class ProxyHTTPConnection(httplib.HTTPConnection):
         port = o.port
         host = o.hostname
         if proto is None:
-            raise ValueError, "unknown URL type: %s" % url
+            raise ValueError("unknown URL type: %s" % url)
         if port is None:
             try:
                 port = self._ports[proto]
             except KeyError:
-                raise ValueError, "unknown protocol for: %s" % url
+                raise ValueError("unknown protocol for: %s" % url)
         self._real_host = host
         self._real_port = port
-        httplib.HTTPConnection.request(self, method, url, body, headers)
+        six.moves.http_client.HTTPConnection.request(self, method, url, body, headers)
 
 
     def connect(self):
-        httplib.HTTPConnection.connect(self)
+        six.moves.http_client.HTTPConnection.connect(self)
         #send proxy CONNECT request
         self.send("CONNECT %s:%d HTTP/1.0\r\n\r\n" % (self._real_host, self._real_port))
         #expect a HTTP/1.0 200 Connection established
@@ -191,7 +194,7 @@ class ProxyHTTPConnection(httplib.HTTPConnection):
         if code != 200:
             #proxy returned and error, abort connection, and raise exception
             self.close()
-            raise socket.error, "Proxy connection failed: %d %s" % (code, message.strip())
+            raise socket.error("Proxy connection failed: %d %s" % (code, message.strip()))
         #eat up header block from proxy....
         while True:
             #should not use directly fp probably
@@ -214,7 +217,7 @@ class ProxyHTTPSConnection(ProxyHTTPConnection):
         #make the sock ssl-aware
         if sys.hexversion < 0x20709f0:
             ssl = socket.ssl(self.sock, self.key_file, self.cert_file)
-            self.sock = httplib.FakeSocket(self.sock, ssl)
+            self.sock = six.moves.http_client.FakeSocket(self.sock, ssl)
         else:
             self.sock = self.context.wrap_socket(self.sock)
 
@@ -223,10 +226,10 @@ class ProxyHTTPSConnection(ProxyHTTPConnection):
 
 class GsiRest(object):
     """docstring for GracedbRest"""
-    def __init__(self, 
-            url=DEFAULT_SERVICE_URL, 
-            proxy_host=None, 
-            proxy_port=3128, 
+    def __init__(self,
+            url=DEFAULT_SERVICE_URL,
+            proxy_host=None,
+            proxy_port=3128,
             cred=None):
         if not cred:
             cred = findUserCredentials()
@@ -234,12 +237,11 @@ class GsiRest(object):
             msg = "\nERROR: No certificate (or proxy) found. \n\n"
             msg += "Please run ligo-proxy-init or grid-proxy-init (as appropriate) "
             msg += "to generate one.\n\n"
-            output_and_die(msg)
+            self.output_and_die(msg)
         if isinstance(cred, (list, tuple)):
             self.cert, self.key = cred
         elif cred:
             self.cert = self.key = cred
-        self.connection = None
 
         o = urlparse(url)
         port = o.port
@@ -255,11 +257,11 @@ class GsiRest(object):
             ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
             try:
                 ssl_context.load_cert_chain(self.cert, self.key)
-            except ssl.SSLError, e:
+            except ssl.SSLError as e:
                 msg = "\nERROR: Unable to load cert/key pair. \n\n"
                 msg += "Please run ligo-proxy-init or grid-proxy-init again "
                 msg += "or make sure your robot certificate is readable.\n\n"
-                output_and_die(msg)
+                self.output_and_die(msg)
             # Generally speaking, test boxes use cheap/free certs from the LIGO CA.
             # These cannot be verified by the client.
             if host in KNOWN_TEST_HOSTS:
@@ -268,19 +270,19 @@ class GsiRest(object):
                 ssl_context.verify_mode = ssl.CERT_REQUIRED
                 ssl_context.check_hostname = True
                 # Find the various CA cert bundles stored on the system
-                ssl_context.load_default_certs()        
+                ssl_context.load_default_certs()
 
             if proxy_host:
                 self.connector = lambda: ProxyHTTPSConnection(proxy_host, proxy_port, context=ssl_context)
             else:
-                self.connector = lambda: httplib.HTTPSConnection(host, port, context=ssl_context)            
+                self.connector = lambda: six.moves.http_client.HTTPSConnection(host, port, context=ssl_context)
         else:
             # Using and older version of python. We'll pass in the cert and key files.
             if proxy_host:
-                self.connector = lambda: ProxyHTTPSConnection(proxy_host, proxy_port, 
+                self.connector = lambda: ProxyHTTPSConnection(proxy_host, proxy_port,
                     key_file=self.key, cert_file=self.cert)
             else:
-                self.connector = lambda: httplib.HTTPSConnection(host, port, 
+                self.connector = lambda: six.moves.http_client.HTTPSConnection(host, port,
                     key_file=self.key, cert_file=self.cert)
 
 
@@ -289,13 +291,13 @@ class GsiRest(object):
 
     # When there is a problem with the SSL connection or cert authentication,
     # either conn.request() or conn.getresponse() will throw an exception.
-    # The following two wrappers are intended to catch these exceptions and 
+    # The following two wrappers are intended to catch these exceptions and
     # return an intelligible error message to the user.
     # A wrapper for getting the response:
     def get_response(self, conn):
         try:
             return conn.getresponse()
-        except ssl.SSLError, e:
+        except ssl.SSLError as e:
             # Check for a valid user proxy cert.
             expired, error = is_expired(self.cert)
             if expired is not None:
@@ -312,24 +314,17 @@ class GsiRest(object):
                 msg = "\nERROR \n\n"
                 msg += "Unable to check certificate expiry date: %s \n\n" %error
                 msg += "Problem establishing secure connection: %s \n\n" % str(e)
-            output_and_die(msg)
-        except Exception, e:
-            # Some error unrelated to the SSL connection has occurred.
-            output_and_die("ERROR: %s\n" % str(e))
+            self.output_and_die(msg)
 
     # A wrapper for making the request.
     def make_request(self, conn, *args, **kwargs):
         try:
             conn.request(*args, **kwargs)
-        except ssl.SSLError, e:
+        except ssl.SSLError as e:
             msg = "\nERROR \n\n"
             msg += "Problem establishing secure connection: %s \n\n" % str(e)
-            output_and_die(msg)
-        except Exception, e:
-            msg = "\nERROR \n\n"
-            msg += "%s \n\n" % str(e)
-            output_and_die(msg)
-        
+            self.output_and_die(msg)
+
     def request(self, method, url, body=None, headers=None, priming_url=None):
         # Bug in Python (versions < 2.7.1 (?))
         # http://bugs.python.org/issue11898
@@ -365,9 +360,9 @@ class GsiRest(object):
                     rdict['retry-after'] = response.getheader('x-throttle-wait-seconds')
                     response_content = json.dumps(rdict)
                 except:
-                    pass  
+                    pass
             raise HTTPError(response.status, response.reason, response_content)
-        response.json = lambda: load_json_or_die(response)
+        response.json = lambda: self.load_json_or_die(response)
         return response
 
     def get(self, url, headers=None):
@@ -400,7 +395,7 @@ class GsiRest(object):
         else:
             body = body or {}
             if isinstance(body, dict):
-                body = body.items()
+                body = list(body.items())
             content_type, body = encode_multipart_formdata(body, files)
 #           XXX What about the headers in the params?
             headers = {
@@ -410,6 +405,37 @@ class GsiRest(object):
             }
         return self.request(method, url, body, headers)
 
+    # A utility for writing out an error message to the user and then stopping
+    # execution. This seems to behave sensibly in both the interpreter and in
+    # a script.
+    @classmethod
+    def output_and_die(cls, msg):
+        raise RuntimeError(msg)
+
+    # Given an HTTPResponse object, try to read its content and interpret as
+    # JSON--or die trying.
+    @classmethod
+    def load_json_or_die(cls, response):
+
+        # First check that the response object actually exists.
+        if not response:
+            raise ValueError("No response object")
+
+        # Next, try to read the content of the response.
+        response_content = response.read()
+        response_content = response_content.decode('utf-8')
+
+        # Finally, try to create a dict by decoding the response as JSON.
+        rdict = None
+        try:
+            rdict = json.loads(response_content)
+        except ValueError:
+            msg = "ERROR: got unexpected content from the server:\n"
+            msg += response_content
+            raise ValueError(msg)
+
+        return rdict
+
 #------------------------------------------------------------------
 # GraceDB
 #
@@ -417,7 +443,7 @@ class GsiRest(object):
 
 class GraceDb(GsiRest):
     """Example GraceDb REST client
-    
+
     The GraceDB service URL may be passed to the constructor
     if an alternate GraceDb instance is desired:
 
@@ -428,7 +454,7 @@ class GraceDb(GsiRest):
     GraceDB behind a proxy. For other kwargs accepted by the constructor,
     consult the source code.
     """
-    def __init__(self, service_url=DEFAULT_SERVICE_URL, 
+    def __init__(self, service_url=DEFAULT_SERVICE_URL,
             proxy_host=None, proxy_port=3128, *args, **kwargs):
         GsiRest.__init__(self, service_url, proxy_host, proxy_port, *args, **kwargs)
 
@@ -453,7 +479,7 @@ class GraceDb(GsiRest):
     @property
     def groups(self):
         return self.service_info.get('groups')
-    
+
     @property
     def pipelines(self):
         return self.service_info.get('pipelines')
@@ -462,6 +488,12 @@ class GraceDb(GsiRest):
     def searches(self):
         return self.service_info.get('searches')
 
+    # Would like to call this 'labels' to keep in line with how
+    # other properties are named, but it's already used for a function.
+    @property
+    def allowed_labels(self):
+        return self.service_info.get('labels')
+
     @property
     def em_groups(self):
         return self.service_info.get('em-groups')
@@ -491,10 +523,10 @@ class GraceDb(GsiRest):
         """Check if input is valid.
            Return coded version if it is"""
         #  code_dict is dict of { code : descriptive_name }
-        if input_value in code_dict.keys():
+        if input_value in list(code_dict.keys()):
             # Already coded
             return input_value
-        if not input_value in code_dict.values():
+        if not input_value in list(code_dict.values()):
             # Not valid
             return None
         return [code
@@ -502,21 +534,27 @@ class GraceDb(GsiRest):
                 if name == input_value][0]
 
     # Search and filecontents are optional when creating an event.
-    def createEvent(self, group, pipeline, filename, search=None, filecontents=None, **kwargs):
+    def createEvent(self, group, pipeline, filename, search=None, labels=None,
+                    offline=False, filecontents=None, **kwargs):
         """Create a new GraceDB event
 
         Required args: group, pipeline, filename
 
-        Optional args: search, filecontents
+        Optional args: search, labels, offline, filecontents
 
-        The filename is the path to a file containing information about the event.
-        The values for 'group', 'pipeline', and 'search' are restricted to those
-        stored in the database.
+        The filename is the path to a file containing information about the
+        event. The values for 'group', 'pipeline', and 'search' are restricted
+        to those stored in the database. 'labels' should be a list of strings
+        corresponding to labels (values restricted to those stored in the
+        database).  'labels' can be a string if only a single label is
+        being applied. 'offline' is a boolean which indicates whether the
+        analysis is offline (True) or online/low-latency (False).
 
         Example:
 
             >>> g = GraceDb()
-            >>> r = g.createEvent('CBC', 'gstlal', '/path/to/something.xml', 'LowMass')
+            >>> r = g.createEvent('CBC', 'gstlal', '/path/to/something.xml',
+            ... labels='INJ', 'LowMass')
             >>> r.status
             201
 
@@ -528,6 +566,28 @@ class GraceDb(GsiRest):
             errors += ["bad pipeline"]
         if search and search not in self.searches:
             errors += ["bad search"]
+        # Process offline arg
+        if not isinstance(offline, bool):
+            errors += ["offline should be True or False"]
+        # Process label args - convert non-empty strings to list
+        # to ensure consistent processing
+        if labels:
+            if isinstance(labels, str):
+                # Convert to list
+                labels = [labels]
+            elif isinstance(labels, list):
+                pass
+            else:
+                # Raise exception instead of adding errors. The next for loop
+                # will break (before errors exception is raised) if labels
+                # is of the wrong type
+                raise TypeError("labels arg is {0}, should be str or list" \
+                    .format(type(labels)))
+            # Check labels against those in database
+            for l in labels:
+                if l not in self.allowed_labels:
+                    raise NameError(("Label '{0}' does not exist in the "
+                        "database").format(l))
         if errors:
             # XXX Terrible error messages / weak exception type
             raise Exception(str(errors))
@@ -536,16 +596,21 @@ class GraceDb(GsiRest):
                 filename = 'initial.data'
                 filecontents = sys.stdin.read()
             else:
-                filecontents = open(filename, 'rb').read() 
+                filecontents = open(filename, 'rb').read()
+
         fields = [
-                  ('group', group),
-                  ('pipeline', pipeline),
-                 ]
+            ('group', group),
+            ('pipeline', pipeline),
+            ('offline', offline),
+        ]
         if search:
             fields.append(('search', search))
+        if labels:
+            for l in labels:
+                fields.append(('labels', l))
 
         # Update fields with additional keyword arguments
-        for key, value in kwargs.iteritems():
+        for key, value in six.iteritems(kwargs):
             fields.append((key, value))
 
         files = [('eventFile', filename, filecontents)]
@@ -559,12 +624,12 @@ class GraceDb(GsiRest):
         Required args: graceid, filename
 
         This function uploads a new event file, hence changing the basic details
-        of an existing event. 
-        
+        of an existing event.
+
         Example:
 
             >>> g = GraceDb()
-            >>> r = g.replaceEvent('T101383', '/path/to/new/something.xml') 
+            >>> r = g.replaceEvent('T101383', '/path/to/new/something.xml')
 
         """
         if filecontents is None:
@@ -578,7 +643,7 @@ class GraceDb(GsiRest):
     def event(self, graceid):
         """Get information about a specific event
 
-        Args: graceid 
+        Args: graceid
 
         Example:
 
@@ -593,8 +658,8 @@ class GraceDb(GsiRest):
         """Get a iterator of events in response to a query
 
         This function returns an iterator which yields event dictionaries.
-        Optional arguments are query, orderby, count, and columns. The 
-        columns argument is a comma separated list of attributes that the 
+        Optional arguments are query, orderby, count, and columns. The
+        columns argument is a comma separated list of attributes that the
         user would like in each event dictionary. If columns are not specified,
         all attributes of the events are returned.
 
@@ -622,7 +687,7 @@ class GraceDb(GsiRest):
                  yield event
 
     def numEvents(self, query=None):
-        """Get the number of events satisfying a query 
+        """Get the number of events satisfying a query
 
         Example:
 
@@ -669,9 +734,9 @@ class GraceDb(GsiRest):
 
         Required args: graceid, filename
 
-        This method creates a new log message with your file attached. It is 
+        This method creates a new log message with your file attached. It is
         strongly preferred to use writeLog() instead of writeFile() so that you
-        can add a more suitable comment. That will make it easier for other 
+        can add a more suitable comment. That will make it easier for other
         users to know what your file contains.
 
         Example:
@@ -701,7 +766,7 @@ class GraceDb(GsiRest):
         Required args: graceid
 
         This function returns a JSON representation of all of an event's
-        log messages. 
+        log messages.
 
         Example:
 
@@ -717,7 +782,7 @@ class GraceDb(GsiRest):
         uri = template.format(graceid=graceid)
         return self.get(uri)
 
-    def writeLog(self, graceid, message, filename=None, filecontents=None, 
+    def writeLog(self, graceid, message, filename=None, filecontents=None,
             tagname=None, displayName=None):
         """Create a new log message
 
@@ -731,8 +796,8 @@ class GraceDb(GsiRest):
         tagname = ["sky_loc", "my_lovely_skymap"]
 
         If only graceid and message are provided, a text comment will be created
-        in the event log. If a filename is also specified, the file will be attached 
-        to the log message and displayed along side the message text. If a tagname 
+        in the event log. If a filename is also specified, the file will be attached
+        to the log message and displayed along side the message text. If a tagname
         is provided, the message will be tagged.
 
         Example:
@@ -757,38 +822,38 @@ class GraceDb(GsiRest):
                 filecontents = filecontents.read()
             files = [('upload', os.path.basename(filename), filecontents)]
 
-        # Let's see if tagname is a string or a list 
-        if tagname and not isinstance(tagname, basestring):
+        # Let's see if tagname is a string or a list
+        if tagname and not isinstance(tagname, six.string_types):
             tagnames = ','.join(tagname)
         else:
             tagnames = tagname if tagname else None
 
-        return self.post(uri, body={'message' : message, 'tagname': tagnames, 
+        return self.post(uri, body={'message' : message, 'tagname': tagnames,
             'displayName': displayName}, files=files)
 
     def eels(self, graceid):
-        """Given a GraceID, get a list of EMBB log entries 
+        """Given a GraceID, get a list of EMBB log entries
 
         Example:
-        
-            >>> g = GraceDb()       
-            >>> r = g.eels('T101383') 
+
+            >>> g = GraceDb()
+            >>> r = g.eels('T101383')
             >>> full_dictionary = r.json()            # Convert the response to a dictionary
             >>> eel_list = full_dictionary['embblog'] # Pull out a list of EEL dicts
 
         """
-    
+
         template = self.templates['embb-event-log-template']
         uri = template.format(graceid=graceid)
         return self.get(uri)
 
-    def writeEel(self, graceid, group, waveband, eel_status, 
+    def writeEel(self, graceid, group, waveband, eel_status,
             obs_status, **kwargs):
-        """Write an EMBB event log entry 
-        
+        """Write an EMBB event log entry
+
         Required args: graceid, group, waveband, eel_status, obs_status
 
-        (Note that 'group' here is the name of the EM MOU group, not 
+        (Note that 'group' here is the name of the EM MOU group, not
         the LVC data analysis group responsible for the original detection.)
 
         Additional keyword arguments may be passed in to be sent in the POST
@@ -807,21 +872,21 @@ class GraceDb(GsiRest):
         wish to convey.
 
         Any other kwargs will be ignored.
-        """ 
+        """
         # validate facility, waveband, eel_status, and obs_status
         if not group in self.em_groups:
             raise ValueError("group must be one of %s" % self.em_groups)
-        
-        if not waveband in self.wavebands.keys():
-            raise ValueError("waveband must be one of %s" % self.wavebands.keys())
+
+        if not waveband in list(self.wavebands.keys()):
+            raise ValueError("waveband must be one of %s" % list(self.wavebands.keys()))
 
         eel_status = self._getCode(eel_status, self.eel_statuses)
         if not eel_status:
-            raise ValueError("EEL status must be one of %s" % self.eel_statuses.values())
+            raise ValueError("EEL status must be one of %s" % list(self.eel_statuses.values()))
 
         obs_status = self._getCode(obs_status, self.obs_statuses)
         if not obs_status:
-            raise ValueError("Observation status must be one of %s" % self.obs_statuses.values())
+            raise ValueError("Observation status must be one of %s" % list(self.obs_statuses.values()))
 
         template = self.templates['embb-event-log-template']
         uri = template.format(graceid=graceid)
@@ -836,35 +901,35 @@ class GraceDb(GsiRest):
         return self.post(uri, body=body)
 
     def emobservations(self, graceid):
-        """Given a GraceID, get a list of EM observation entries 
+        """Given a GraceID, get a list of EM observation entries
 
         Example:
-        
-            >>> g = GraceDb()       
-            >>> r = g.emobserations('T101383') 
+
+            >>> g = GraceDb()
+            >>> r = g.emobservations('T101383')
             >>> full_dictionary = r.json()            # Convert the response to a dictionary
             >>> emo_list = full_dictionary['observations'] # Pull out a list of EMO dicts
 
         """
-    
+
         template = self.templates['emobservation-list-template']
         uri = template.format(graceid=graceid)
         return self.get(uri)
 
     def writeEMObservation(self, graceid, group, raList, raWidthList,
         decList, decWidthList, startTimeList, durationList, comment=None):
-        """Write an EM observation entry 
-        
+        """Write an EM observation entry
+
         Required args: graceid, group, raList, decList, raWidthList,
             decWidthList, startTimeList, durationList
 
-        The various lists arguments should contain Python lists or 
-        comma-separated values (or a single value). Start times are 
+        The various lists arguments should contain Python lists or
+        comma-separated values (or a single value). Start times are
         in ISO 8601 UTC. Durations are in seconds.
 
-        (Note that 'group' here is the name of the EM MOU group, not 
+        (Note that 'group' here is the name of the EM MOU group, not
         the LVC data analysis group responsible for the original detection.)
-        """ 
+        """
         # validate facility, waveband, eel_status, and obs_status
         if not group in self.em_groups:
             raise ValueError("group must be one of %s" % self.em_groups)
@@ -872,23 +937,23 @@ class GraceDb(GsiRest):
         # NOTE: One could do validation of the various list inputs here
         # rather than relying on the server.
         # These arguments can consist of a single element, a python
-        # list, or a string.  Transform the list args to csv (unless they 
+        # list, or a string.  Transform the list args to csv (unless they
         # already are)
-        raList, raWidthList, decList, decWidthList, startTimeList, durationList = map(
-            cleanListInput, 
-            [raList, raWidthList, decList, decWidthList, startTimeList, durationList])
+        raList, raWidthList, decList, decWidthList, startTimeList, durationList = list(map(
+            cleanListInput,
+            [raList, raWidthList, decList, decWidthList, startTimeList, durationList]))
 
         template = self.templates['emobservation-list-template']
         uri = template.format(graceid=graceid)
 
-        body = { 
-            'group' : group, 
+        body = {
+            'group' : group,
             'raList' : raList,
             'raWidthList': raWidthList,
             'decList' : decList,
             'decWidthList': decWidthList,
             'startTimeList': startTimeList,
-            'durationList': durationList, 
+            'durationList': durationList,
             'comment': comment,
         }
         return self.post(uri, body=body)
@@ -914,7 +979,7 @@ class GraceDb(GsiRest):
 
         Required args:  graceid, label name
 
-        The label name must correspond to one of the existing 
+        The label name must correspond to one of the existing
         GraceDB labels, else an error will result.
 
         Example:
@@ -972,7 +1037,7 @@ class GraceDb(GsiRest):
 
         Required arguments: graceid, n (the number of the log message)
         and the tagname. If a displayName is provided (and if the tag
-        doesn't already exist), a new tag will be created with the 
+        doesn't already exist), a new tag will be created with the
         provided display name.
 
         Example:
@@ -991,10 +1056,10 @@ class GraceDb(GsiRest):
         """Remove a tag from a given log message
 
         Required arguments: graceid, n (the number of the log message)
-        and the tagname.        
+        and the tagname.
 
         Example:
-            
+
             >>> g = GraceDb()
             >>> r = g.deleteTag('T101383', 56, 'sky_loc')
             >>> r.status
@@ -1025,20 +1090,20 @@ class GraceDb(GsiRest):
         """Given a GraceID, get a list of VOEvents
 
         Example:
-        
-            >>> g = GraceDb()       
-            >>> r = g.voevents('T101383') 
-            >>> voevent_list = r.json()['voevents'] 
+
+            >>> g = GraceDb()
+            >>> r = g.voevents('T101383')
+            >>> voevent_list = r.json()['voevents']
 
         """
-    
+
         template = self.templates['voevent-list-template']
         uri = template.format(graceid=graceid)
         return self.get(uri)
 
     def createVOEvent(self, graceid, voevent_type, **kwargs):
         """Create a new VOEvent
-        
+
         Required args: graceid, voevent_type
 
         Additional keyword arguments may be passed in to be sent in the POST
@@ -1047,11 +1112,11 @@ class GraceDb(GsiRest):
             skyamp_type
 
         Any other kwargs will be ignored.
-        """ 
+        """
         # validate facility, waveband, eel_status, and obs_status
         voevent_type = self._getCode(voevent_type.lower(), self.voevent_types)
         if not voevent_type:
-            raise ValueError("voevent_type must be one of: %s" % self.voevent_types.values())
+            raise ValueError("voevent_type must be one of: %s" % list(self.voevent_types.values()))
         template = self.templates['voevent-list-template']
         uri = template.format(graceid=graceid)
 
@@ -1083,7 +1148,7 @@ class GraceDb(GsiRest):
 
 class GraceDbBasic(GraceDb):
     """Example GraceDb REST client with basic auth
-    
+
     The GraceDB service URL may be passed to the constructor
     if an alternate GraceDb instance is desired:
 
@@ -1095,7 +1160,7 @@ class GraceDbBasic(GraceDb):
     consult the source code.
     """
     def __init__(self, service_url=DEFAULT_BASIC_SERVICE_URL,
-            proxy_host=None, proxy_port=3128, username=None, password=None, 
+            proxy_host=None, proxy_port=3128, username=None, password=None,
             *args, **kwargs):
 
         o = urlparse(service_url)
@@ -1105,14 +1170,14 @@ class GraceDbBasic(GraceDb):
 
         if not username or not password:
             try:
-                username, account, password = netrc.netrc().authenticators(host)
-            except:
-                pass 
+                username, account, password = safe_netrc().authenticators(host)
+            except IOError:
+                pass
 
         if not username or not password:
             msg = "\nERROR: Could not find user credentials. \n\n"
-            msg +="Please use a .netrc file or provide username and password.\n\n" 
-            output_and_die(msg)
+            msg +="Please use a .netrc file or provide username and password.\n\n"
+            self.output_and_die(msg)
 
         # Construct authorization header
         userAndPass = b64encode(b"%s:%s" % (username, password)).decode("ascii")
@@ -1133,18 +1198,18 @@ class GraceDbBasic(GraceDb):
                 ssl_context.verify_mode = ssl.CERT_REQUIRED
                 ssl_context.check_hostname = True
                 # Find the various CA cert bundles stored on the system
-                ssl_context.load_default_certs()        
+                ssl_context.load_default_certs()
 
             if proxy_host:
                 self.connector = lambda: ProxyHTTPSConnection(proxy_host, proxy_port, context=ssl_context)
             else:
-                self.connector = lambda: httplib.HTTPSConnection(host, port, context=ssl_context)            
+                self.connector = lambda: six.moves.http_client.HTTPSConnection(host, port, context=ssl_context)
         else:
             # Using and older version of python. We'll pass in the cert and key files.
             if proxy_host:
                 self.connector = lambda: ProxyHTTPSConnection(proxy_host, proxy_port)
             else:
-                self.connector = lambda: httplib.HTTPSConnection(host, port)
+                self.connector = lambda: six.moves.http_client.HTTPSConnection(host, port)
 
         self.service_url = service_url
         self._service_info = None
@@ -1154,27 +1219,10 @@ class GraceDbBasic(GraceDb):
     def get_response(self, conn):
         try:
             return conn.getresponse()
-        except ssl.SSLError, e:
+        except ssl.SSLError as e:
             msg = "\nERROR \n\n"
             msg += "Problem establishing secure connection: %s \n\n" % str(e)
-            output_and_die(msg)
-        except Exception, e:
-            msg = "\nERROR \n\n"
-            msg += "%s \n\n" % str(e)
-            output_and_die(msg)
-
-    # A wrapper for making the request.
-    def make_request(self, conn, *args, **kwargs):
-        try:
-            conn.request(*args, **kwargs)
-        except ssl.SSLError, e:
-            msg = "\nERROR \n\n"
-            msg += "Problem establishing secure connection: %s \n\n" % str(e)
-            output_and_die(msg)
-        except Exception, e:
-            msg = "\nERROR \n\n"
-            msg += "%s \n\n" % str(e)
-            output_and_die(msg)
+            self.output_and_die(msg)
 
     def request(self, method, url, body=None, headers=None):
         # Bug in Python (versions < 2.7.1 (?))
@@ -1197,11 +1245,11 @@ class GraceDbBasic(GraceDb):
             try:
                 msg = "\nERROR: %s \n\n" % json.loads(response.read())['error']
             except:
-                msg = "\nERROR: \n\n" 
+                msg = "\nERROR: \n\n"
             msg += "Please check the username/password in your .netrc file. \n"
             msg += "Note: If your password is more than a year old, you will \n"
-            msg += "need to use the web interface to generate a new one. \n\n" 
-            output_and_die(msg)
+            msg += "need to use the web interface to generate a new one. \n\n"
+            self.output_and_die(msg)
         return self.adjustResponse(response)
 
 #-----------------------------------------------------------------
